from .pairing_heap cimport MinPairingHeap
from .event cimport Event
from .sim cimport Environment


cdef class Process:
    cdef readonly Environment env
    # cdef readonly double time
    cdef object process
    cdef object id
    cdef readonly bint _is_waiting
    cdef Event event
    cdef MinPairingHeap ev_heap

    cpdef tuple _wait(self)
    cpdef _process(self)
    cpdef void activate(self, double time, int reason)
    cpdef void start(self) except *
    cpdef void finish(self)
    cpdef void setup_env(self, env)
    cdef void run_step(self) except *





from libc.stdint cimport int64_t

cpdef int64_t d2l(double x)
cpdef double l2d(int64_t x)

cpdef bint time_eq(double a, double b)
cpdef bint time_le(double a, double b)
cpdef bint time_lt(double a, double b)
cpdef bint time_ge(double a, double b)
cpdef bint time_gt(double a, double b)



import sys, os

sys.path = [os.path.realpath("../build/lib.linux-x86_64-3.8")] + sys.path

from enum import IntEnum, auto

from pesim import TIME_REACHED, make_process, Environment, TIME_FOREVER, TIME_PASSED
from pesim.lock import Lock, Semaphore, WaitEvent
from collections import deque
from random import randint

task_queue = deque()


class Reason(IntEnum):
    NewTask = auto()
    TaskAssign = auto()
    TaskDone = auto()


lock = Lock()


def proc_a(self):
    yield self.time + 10, TIME_PASSED
    yield lock.acquire(self)
    print("a", self.time, "acquired")
    yield self.time + 10, TIME_PASSED
    print("a", self.time, "to release")
    lock.release()


def proc_b(self):
    yield lock.acquire(self)
    yield self.time + 15, TIME_PASSED
    print("b", self.time, "to release")
    lock.release()
    yield lock.acquire(self)
    print("b", self.time, "acquired")


if __name__ == '__main__':
    env = Environment()

    a = make_process(env, proc_a)
    b = make_process(env, proc_b)

    env.start()
    env.run_until(TIME_FOREVER - 1, TIME_PASSED)



from .process cimport Process
from .pairing_heap cimport MinPairingHeap
from .event cimport Event
from libc.stdint cimport int64_t

cdef class Environment:
    cdef MinPairingHeap ev_heap
    cdef int64_t time_i64
    cdef list processes
    cdef readonly bint started
    cdef readonly double stop_time

    cpdef void add(self, Process process)
    cdef inline void timeout(self, Event ev, double time, int reason)
    cpdef void start(self) except *
    cpdef void finish(self)
    cpdef double run_until(self, double ex_time, int after_reason=?)
    cpdef double join(self)
    cpdef double next_event_time(self)
    cpdef Event next_event(self)



from cpython cimport PyObject


cdef class MinPairingHeapNode:
    cdef PyObject *first_child
    cdef PyObject *left
    cdef PyObject *right

    cpdef bint key_lt(self, MinPairingHeapNode other)
    cdef void _insert_first_child(self, PyObject*node)
    cdef void _detach(self)
    cdef PyObject*_meld(self, PyObject*other)
    cdef PyObject*_pop(self)
    cpdef void print_tree(self, int indent=?)
    cdef clear_subtree(self)

cdef class MinPairingHeap:
    cdef PyObject*root

    cpdef MinPairingHeapNode first(self)
    cpdef void push(self, MinPairingHeapNode node)
    cdef void push_x(self, PyObject* ptr)
    cpdef MinPairingHeapNode pop(self)
    cpdef void notify_dec(self, MinPairingHeapNode node)
    cdef void notify_dec_x(self, PyObject* ptr)
    cpdef clear(self)



from setuptools import setup, Extension, find_packages
from platform import system
from Cython.Build import cythonize


if system() == "Windows":
    extra_compile_args = [
        "/fp:fast",
        # "/arch:AVX512",
        "/favor:INTEL64",
        "/O2", "/Ob2", "/GL",
        ]
    extra_link_args = [
        "/MACHINE:X64",
        "/LTCG",
        ]
    library_dir = "libtcy/msvc/Release"
else:
    extra_compile_args = [
        '-Ofast',
        '-march=native',
        '-ffast-math',
        '-fforce-addr',
        '-fprefetch-loop-arrays',
        "-flto",
        ]
    extra_link_args = []
    library_dir = "libtcy/cmake-build-debug"

extensions = [
    Extension(
        name="pesim.*",
        sources=["pesim/*.pyx"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        ),
    ]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="pesim",
    packages=find_packages(),
    version="0.9.5.0",
    setup_requires=["Cython"],
    author="Tefx",
    author_email="zhaomeng.zhu@gmail.com",
    url="https://github.com/Tefx/pesim",
    project_urls={
        'Documentation': "https://pesim.readthedocs.io",
        "Source Code": "https://github.com/Tefx/pesim",
        },
    description="A Minimalist Discrete Event Simulation Library in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        ],
    ext_modules=cythonize(
        extensions,
        compiler_directives={
            # "embedsignature":   True,
            "profile":          False,
            "linetrace":        False,
            "cdivision":        True,
            "boundscheck":      False,
            "wraparound":       False,
            "initializedcheck": False,
            "language_level":   3,
            },
        annotate=True,
        ),
    package_data={
        "":["*.pxd"],
        },
    )



from .process cimport Process
from cpython cimport PyObject


cdef class _SyncObj:
    cdef PyObject** wait_processes
    cdef int wait_max
    cdef int num_wait
    cdef readonly int value
    cdef readonly dict args

    cdef bint _async_acquire(self, Process proc, int change)
    cdef tuple _acquire(self, Process proc, int change)
    cdef void _release(self, int change, int acquire_change, int reason)

cdef class Semaphore(_SyncObj):
    pass

cdef class Lock(_SyncObj):
    pass

cdef class Condition(_SyncObj):
    pass

cdef class RLock:
    cdef PyObject* holder
    cdef object wait_objs
    cdef readonly int value
    cdef readonly dict args

    cdef bint _async_acquire(self, Process proc, object obj)
    cdef tuple _acquire(self, Process proc, object obj)





from .process cimport Process
from .pairing_heap cimport MinPairingHeapNode
from libc.stdint cimport int64_t

cdef class Event(MinPairingHeapNode):
    cdef int64_t time_i64
    cdef Process process
    cdef int reason

    cpdef bint key_lt(self, MinPairingHeapNode other)



from .sim import Environment
from .define import TIME_FOREVER, TIME_PASSED, TIME_REACHED
from .process import Process
from .pairing_heap import MinPairingHeap, MinPairingHeapNode
from .lock import Lock, RLock, Semaphore, Condition
